from enum import StrEnum


class StandardIcon(StrEnum):
    ADVANCED = "advanced"
    ANGLE_DOWN = "angleDown"
    ANGLE_LEFT = "angleLeft"
    ANGLE_RIGHT = "angleRight"
    ANGLES_DOWN = "anglesDown"
    ANGLES_LEFT = "anglesLeft"
    ANGLES_RIGHT = "anglesRight"
    ANGLES_UP = "anglesUp"
    ANGLE_UP = "angleUp"
    ARROW_DOWN = "arrowDown"
    ARROW_LEFT = "arrowLeft"
    ARROW_RIGHT = "arrowRight"
    ARROW_UP = "arrowUp"
    AWARD = "award"
    BACKGROUND = "background"
    BELL = "bell"
    BELL_SLASH = "bellSlash"
    BOOKMARK = "bookmark"
    BOOKMARK_HOLLOW = "bookmarkHollow"
    CHECK = "check"
    CLOCK = "clock"
    COMMENT = "comment"
    CROWN = "crown"
    DELETE = "delete"
    EDIT = "edit"
    EFFECT = "effect"
    ENGINE = "engine"
    ENVELOPE = "envelope"
    ENVELOPE_OPEN = "envelopeOpen"
    GLOBE = "globe"
    HEART = "heart"
    HEART_HOLLOW = "heartHollow"
    HIDE = "hide"
    INFORMATION = "information"
    LEVEL = "level"
    LOCK = "lock"
    MEDAL = "medal"
    MESSAGE = "message"
    MINUS = "minus"
    OPTIONS = "options"
    PARTICLE = "particle"
    PIN = "pin"
    PLAYER = "player"
    PLAYLIST = "playlist"
    PLUS = "plus"
    POST = "post"
    QUESTION = "question"
    RANKING = "ranking"
    REPLAY = "replay"
    REPLY = "reply"
    RESTORE = "restore"
    ROOM = "room"
    SEARCH = "search"
    SETTINGS = "settings"
    SHOW = "show"
    SHUFFLE = "shuffle"
    SKIN = "skin"
    STAR = "star"
    STAR_HALF = "starHalf"
    STAR_HOLLOW = "starHollow"
    STOPWATCH = "stopwatch"
    TAG = "tag"
    THUMBS_DOWN = "thumbsDown"
    THUMBS_DOWN_HOLLOW = "thumbsDownHollow"
    THUMBS_UP = "thumbsUp"
    THUMBS_UP_HOLLOW = "thumbsUpHollow"
    TROPHY = "trophy"
    UNLOCK = "unlock"
    X_MARK = "xMark"
