EMPTY_ENGINE_PLAY_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "buckets": [],
    "archetypes": [],
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_WATCH_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "buckets": [],
    "archetypes": [],
    "updateSpawn": 0,
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_PREVIEW_DATA = {
    "skin": {"sprites": []},
    "archetypes": [],
    "nodes": [{"value": 0}],
}

EMPTY_ENGINE_TUTORIAL_DATA = {
    "skin": {"sprites": []},
    "effect": {"clips": []},
    "particle": {"effects": []},
    "instruction": {"texts": [], "icons": []},
    "nodes": [{"value": 0}],
}
