class StandardText:
    Post = "#POST"
    Playlist = "#PLAYLIST"
    Level = "#LEVEL"
    Skin = "#SKIN"
    Background = "#BACKGROUND"
    Effect = "#EFFECT"
    Particle = "#PARTICLE"
    Engine = "#ENGINE"
    Replay = "#REPLAY"
    Room = "#ROOM"
    Grade = "#GRADE"
    ArcadeScore = "#ARCADE_SCORE"
    AccuracyScore = "#ACCURACY_SCORE"
    Combo = "#COMBO"
    Perfect = "#PERFECT"
    Great = "#GREAT"
    Good = "#GOOD"
    Miss = "#MISS"
    Keywords = "#KEYWORDS"
    Name = "#NAME"
    Rating = "#RATING"
    RatingMinimum = "#RATING_MINIMUM"
    RatingMaximum = "#RATING_MAXIMUM"
    Title = "#TITLE"
    Subtitle = "#SUBTITLE"
    Artists = "#ARTISTS"
    Time = "#TIME"
    Author = "#AUTHOR"
    Genre = "#GENRE"
    Type = "#TYPE"
    Category = "#CATEGORY"
    Language = "#LANGUAGE"
    Difficulty = "#DIFFICULTY"
    Version = "#VERSION"
    Length = "#LENGTH"
    LengthMinimum = "#LENGTH_MINIMUM"
    LengthMaximum = "#LENGTH_MAXIMUM"
    AdditionalInformation = "#ADDITIONAL_INFORMATION"
    Timezone = "#TIMEZONE"
    Region = "#REGION"
    Content = "#CONTENT"
    Comment = "#COMMENT"
    Review = "#REVIEW"
    Reply = "#REPLY"
    Message = "#MESSAGE"
    Role = "#ROLE"
    Permission = "#PERMISSION"
    User = "#USER"
    Speed = "#SPEED"
    Mirror = "#MIRROR"
    Random = "#RANDOM"
    Hidden = "#HIDDEN"
    JudgmentStrict = "#JUDGMENT_STRICT"
    JudgmentLoose = "#JUDGMENT_LOOSE"
    EffectAuto = "#EFFECT_AUTO"
    Stage = "#STAGE"
    StagePosition = "#STAGE_POSITION"
    StageSize = "#STAGE_SIZE"
    StageRotation = "#STAGE_ROTATION"
    StageDirection = "#STAGE_DIRECTION"
    StageAlpha = "#STAGE_ALPHA"
    StageAnimation = "#STAGE_ANIMATION"
    StageTilt = "#STAGE_TILT"
    StageCoverVertical = "#STAGE_COVER_VERTICAL"
    StageCoverHorizontal = "#STAGE_COVER_HORIZONTAL"
    StageCoverAlpha = "#STAGE_COVER_ALPHA"
    StageAspectratioLock = "#STAGE_ASPECTRATIO_LOCK"
    StageEffect = "#STAGE_EFFECT"
    StageEffectPosition = "#STAGE_EFFECT_POSITION"
    StageEffectSize = "#STAGE_EFFECT_SIZE"
    StageEffectRotation = "#STAGE_EFFECT_ROTATION"
    StageEffectDirection = "#STAGE_EFFECT_DIRECTION"
    StageEffectAlpha = "#STAGE_EFFECT_ALPHA"
    Lane = "#LANE"
    LanePosition = "#LANE_POSITION"
    LaneSize = "#LANE_SIZE"
    LaneRotation = "#LANE_ROTATION"
    LaneDirection = "#LANE_DIRECTION"
    LaneAlpha = "#LANE_ALPHA"
    LaneAnimation = "#LANE_ANIMATION"
    LaneEffect = "#LANE_EFFECT"
    LaneEffectPosition = "#LANE_EFFECT_POSITION"
    LaneEffectSize = "#LANE_EFFECT_SIZE"
    LaneEffectRotation = "#LANE_EFFECT_ROTATION"
    LaneEffectDirection = "#LANE_EFFECT_DIRECTION"
    LaneEffectAlpha = "#LANE_EFFECT_ALPHA"
    Judgeline = "#JUDGELINE"
    JudgelinePosition = "#JUDGELINE_POSITION"
    JudgelineSize = "#JUDGELINE_SIZE"
    JudgelineRotation = "#JUDGELINE_ROTATION"
    JudgelineDirection = "#JUDGELINE_DIRECTION"
    JudgelineAlpha = "#JUDGELINE_ALPHA"
    JudgelineAnimation = "#JUDGELINE_ANIMATION"
    JudgelineEffect = "#JUDGELINE_EFFECT"
    JudgelineEffectPosition = "#JUDGELINE_EFFECT_POSITION"
    JudgelineEffectSize = "#JUDGELINE_EFFECT_SIZE"
    JudgelineEffectRotation = "#JUDGELINE_EFFECT_ROTATION"
    JudgelineEffectDirection = "#JUDGELINE_EFFECT_DIRECTION"
    JudgelineEffectAlpha = "#JUDGELINE_EFFECT_ALPHA"
    Slot = "#SLOT"
    SlotPosition = "#SLOT_POSITION"
    SlotSize = "#SLOT_SIZE"
    SlotRotation = "#SLOT_ROTATION"
    SlotDirection = "#SLOT_DIRECTION"
    SlotAlpha = "#SLOT_ALPHA"
    SlotAnimation = "#SLOT_ANIMATION"
    SlotEffect = "#SLOT_EFFECT"
    SlotEffectPosition = "#SLOT_EFFECT_POSITION"
    SlotEffectSize = "#SLOT_EFFECT_SIZE"
    SlotEffectRotation = "#SLOT_EFFECT_ROTATION"
    SlotEffectDirection = "#SLOT_EFFECT_DIRECTION"
    SlotEffectAlpha = "#SLOT_EFFECT_ALPHA"
    Note = "#NOTE"
    NoteSpeed = "#NOTE_SPEED"
    NoteSpeedRandom = "#NOTE_SPEED_RANDOM"
    NotePosition = "#NOTE_POSITION"
    NoteSize = "#NOTE_SIZE"
    NoteRotation = "#NOTE_ROTATION"
    NoteDirection = "#NOTE_DIRECTION"
    NoteColor = "#NOTE_COLOR"
    NoteAlpha = "#NOTE_ALPHA"
    NoteAnimation = "#NOTE_ANIMATION"
    NoteEffect = "#NOTE_EFFECT"
    NoteEffectPosition = "#NOTE_EFFECT_POSITION"
    NoteEffectSize = "#NOTE_EFFECT_SIZE"
    NoteEffectRotation = "#NOTE_EFFECT_ROTATION"
    NoteEffectDirection = "#NOTE_EFFECT_DIRECTION"
    NoteEffectColor = "#NOTE_EFFECT_COLOR"
    NoteEffectAlpha = "#NOTE_EFFECT_ALPHA"
    Marker = "#MARKER"
    MarkerPosition = "#MARKER_POSITION"
    MarkerSize = "#MARKER_SIZE"
    MarkerRotation = "#MARKER_ROTATION"
    MarkerDirection = "#MARKER_DIRECTION"
    MarkerColor = "#MARKER_COLOR"
    MarkerAlpha = "#MARKER_ALPHA"
    MarkerAnimation = "#MARKER_ANIMATION"
    Connector = "#CONNECTOR"
    ConnectorPosition = "#CONNECTOR_POSITION"
    ConnectorSize = "#CONNECTOR_SIZE"
    ConnectorRotation = "#CONNECTOR_ROTATION"
    ConnectorDirection = "#CONNECTOR_DIRECTION"
    ConnectorColor = "#CONNECTOR_COLOR"
    ConnectorAlpha = "#CONNECTOR_ALPHA"
    ConnectorAnimation = "#CONNECTOR_ANIMATION"
    Simline = "#SIMLINE"
    SimlinePosition = "#SIMLINE_POSITION"
    SimlineSize = "#SIMLINE_SIZE"
    SimlineRotation = "#SIMLINE_ROTATION"
    SimlineDirection = "#SIMLINE_DIRECTION"
    SimlineColor = "#SIMLINE_COLOR"
    SimlineAlpha = "#SIMLINE_ALPHA"
    SimlineAnimation = "#SIMLINE_ANIMATION"
    None_ = "#NONE"
    Any = "#ANY"
    All = "#ALL"
    Others = "#OTHERS"
    Short = "#SHORT"
    Long = "#LONG"
    High = "#HIGH"
    Mid = "#MID"
    Low = "#LOW"
    Small = "#SMALL"
    Medium = "#MEDIUM"
    Large = "#LARGE"
    Left = "#LEFT"
    Right = "#RIGHT"
    Up = "#UP"
    Down = "#DOWN"
    Front = "#FRONT"
    Back = "#BACK"
    Center = "#CENTER"
    Top = "#TOP"
    Bottom = "#BOTTOM"
    TopLeft = "#TOP_LEFT"
    TopCenter = "#TOP_CENTER"
    TopRight = "#TOP_RIGHT"
    CenterLeft = "#CENTER_LEFT"
    CenterRight = "#CENTER_RIGHT"
    BottomLeft = "#BOTTOM_LEFT"
    BottomCenter = "#BOTTOM_CENTER"
    BottomRight = "#BOTTOM_RIGHT"
    Clockwise = "#CLOCKWISE"
    Counterclockwise = "#COUNTERCLOCKWISE"
    Forward = "#FORWARD"
    Backward = "#BACKWARD"
    Default = "#DEFAULT"
    Neutral = "#NEUTRAL"
    Red = "#RED"
    Green = "#GREEN"
    Blue = "#BLUE"
    Yellow = "#YELLOW"
    Purple = "#PURPLE"
    Cyan = "#CYAN"
    Simple = "#SIMPLE"
    Easy = "#EASY"
    Normal = "#NORMAL"
    Hard = "#HARD"
    Expert = "#EXPERT"
    Master = "#MASTER"
    Pro = "#PRO"
    Technical = "#TECHNICAL"
    Special = "#SPECIAL"
    KeywordsPlaceholder = "#KEYWORDS_PLACEHOLDER"
    NamePlaceholder = "#NAME_PLACEHOLDER"
    RatingPlaceholder = "#RATING_PLACEHOLDER"
    RatingMinimumPlaceholder = "#RATING_MINIMUM_PLACEHOLDER"
    RatingMaximumPlaceholder = "#RATING_MAXIMUM_PLACEHOLDER"
    TitlePlaceholder = "#TITLE_PLACEHOLDER"
    SubtitlePlaceholder = "#SUBTITLE_PLACEHOLDER"
    ArtistsPlaceholder = "#ARTISTS_PLACEHOLDER"
    TimePlaceholder = "#TIME_PLACEHOLDER"
    AuthorPlaceholder = "#AUTHOR_PLACEHOLDER"
    GenrePlaceholder = "#GENRE_PLACEHOLDER"
    TypePlaceholder = "#TYPE_PLACEHOLDER"
    CategoryPlaceholder = "#CATEGORY_PLACEHOLDER"
    LanguagePlaceholder = "#LANGUAGE_PLACEHOLDER"
    DifficultyPlaceholder = "#DIFFICULTY_PLACEHOLDER"
    LengthPlaceholder = "#LENGTH_PLACEHOLDER"
    LengthMinimumPlaceholder = "#LENGTH_MINIMUM_PLACEHOLDER"
    LengthMaximumPlaceholder = "#LENGTH_MAXIMUM_PLACEHOLDER"
    AdditionalInformationPlaceholder = "#ADDITIONAL_INFORMATION_PLACEHOLDER"
    TimezonePlaceholder = "#TIMEZONE_PLACEHOLDER"
    RegionPlaceholder = "#REGION_PLACEHOLDER"
    ContentPlaceholder = "#CONTENT_PLACEHOLDER"
    CommentPlaceholder = "#COMMENT_PLACEHOLDER"
    ReviewPlaceholder = "#REVIEW_PLACEHOLDER"
    ReplyPlaceholder = "#REPLY_PLACEHOLDER"
    MessagePlaceholder = "#MESSAGE_PLACEHOLDER"
    RolePlaceholder = "#ROLE_PLACEHOLDER"
    PermissionPlaceholder = "#PERMISSION_PLACEHOLDER"
    UserPlaceholder = "#USER_PLACEHOLDER"
    PercentageUnit = "#PERCENTAGE_UNIT"
    SecondUnit = "#SECOND_UNIT"
    MillisecondUnit = "#MILLISECOND_UNIT"
    Tap = "#TAP"
    TapHold = "#TAP_HOLD"
    TapRelease = "#TAP_RELEASE"
    TapFlick = "#TAP_FLICK"
    TapSlide = "#TAP_SLIDE"
    Hold = "#HOLD"
    HoldSlide = "#HOLD_SLIDE"
    HoldFollow = "#HOLD_FOLLOW"
    Release = "#RELEASE"
    Flick = "#FLICK"
    Slide = "#SLIDE"
    SlideFlick = "#SLIDE_FLICK"
    Avoid = "#AVOID"
    Jiggle = "#JIGGLE"
    Newest = "#NEWEST"
    Oldest = "#OLDEST"
    Recommended = "#RECOMMENDED"
    Popular = "#POPULAR"
    Featured = "#FEATURED"
    Competitive = "#COMPETITIVE"
    Holiday = "#HOLIDAY"
    Limited = "#LIMITED"
    Following = "#FOLLOWING"
    Announcement = "#ANNOUNCEMENT"
    Information = "#INFORMATION"
    Help = "#HELP"
    Maintenance = "#MAINTENANCE"
    Event = "#EVENT"
    Update = "#UPDATE"
    Search = "#SEARCH"
    Advanced = "#ADVANCED"
    Related = "#RELATED"
    SameAuthor = "#SAME_AUTHOR"
    SameArtists = "#SAME_ARTISTS"
    SameRating = "#SAME_RATING"
    SameCategory = "#SAME_CATEGORY"
    SameDifficulty = "#SAME_DIFFICULTY"
    SameGenre = "#SAME_GENRE"
    SameVersion = "#SAME_VERSION"
    OtherAuthors = "#OTHER_AUTHORS"
    OtherArtists = "#OTHER_ARTISTS"
    OtherRatings = "#OTHER_RATINGS"
    OtherCategories = "#OTHER_CATEGORIES"
    OtherDifficulties = "#OTHER_DIFFICULTIES"
    OtherGenres = "#OTHER_GENRES"
    OtherVersions = "#OTHER_VERSIONS"
    Draft = "#DRAFT"
    Reviewing = "#REVIEWING"
    Published = "#PUBLISHED"
    Verifying = "#VERIFYING"
    Verified = "#VERIFIED"
    Removed = "#REMOVED"
    Public = "#PUBLIC"
    Private = "#PRIVATE"
    Locked = "#LOCKED"
    Pop = "#POP"
    Rock = "#ROCK"
    Hiphop = "#HIPHOP"
    Country = "#COUNTRY"
    Electronic = "#ELECTRONIC"
    Metal = "#METAL"
    Classical = "#CLASSICAL"
    Folk = "#FOLK"
    Indie = "#INDIE"
    Anime = "#ANIME"
    Vocaloid = "#VOCALOID"
    Remix = "#REMIX"
    Instrumental = "#INSTRUMENTAL"
    ShortVersion = "#SHORT_VERSION"
    LongVersion = "#LONG_VERSION"
    LiveVersion = "#LIVE_VERSION"
    Report = "#REPORT"
    Reason = "#REASON"
    IllegalActivities = "#ILLEGAL_ACTIVITIES"
    Cheating = "#CHEATING"
    Afk = "#AFK"
    Spamming = "#SPAMMING"
    VerbalAbuse = "#VERBAL_ABUSE"
    InappropriateLanguage = "#INAPPROPRIATE_LANGUAGE"
    NegativeAttitude = "#NEGATIVE_ATTITUDE"
    Dnf = "#DNF"
    Suggestions = "#SUGGESTIONS"
    SuggestionsPerPlayer = "#SUGGESTIONS_PER_PLAYER"
    MatchScoring = "#MATCH_SCORING"
    MatchTiebreaker = "#MATCH_TIEBREAKER"
    MatchCount = "#MATCH_COUNT"
    MatchLimit = "#MATCH_LIMIT"
    RoundScoring = "#ROUND_SCORING"
    RoundTiebreaker = "#ROUND_TIEBREAKER"
    RoundCount = "#ROUND_COUNT"
    RoundLimit = "#ROUND_LIMIT"
    TeamScoring = "#TEAM_SCORING"
    TeamTiebreaker = "#TEAM_TIEBREAKER"
    TeamCount = "#TEAM_COUNT"
    TeamLimit = "#TEAM_LIMIT"
    Qualified = "#QUALIFIED"
    Disqualified = "#DISQUALIFIED"
    Ranking = "#RANKING"
    Score = "#SCORE"
    Player = "#PLAYER"
    Spectator = "#SPECTATOR"
    Referee = "#REFEREE"
    Eliminated = "#ELIMINATED"
    Finalist = "#FINALIST"
    Finished = "#FINISHED"
    Winner = "#WINNER"
    GoldMedal = "#GOLD_MEDAL"
    SilverMedal = "#SILVER_MEDAL"
    BronzeMedal = "#BRONZE_MEDAL"
    Team1 = "#TEAM_1"
    Team2 = "#TEAM_2"
    Team3 = "#TEAM_3"
    Team4 = "#TEAM_4"
    Team5 = "#TEAM_5"
    Team6 = "#TEAM_6"
    Team7 = "#TEAM_7"
    Team8 = "#TEAM_8"
    TeamRed = "#TEAM_RED"
    TeamGreen = "#TEAM_GREEN"
    TeamBlue = "#TEAM_BLUE"
    TeamYellow = "#TEAM_YELLOW"
    TeamPurple = "#TEAM_PURPLE"
    TeamCyan = "#TEAM_CYAN"
    TeamWhite = "#TEAM_WHITE"
    TeamBlack = "#TEAM_BLACK"
    Upload = "#UPLOAD"
    Submit = "#SUBMIT"
    Edit = "#EDIT"
    Like = "#LIKE"
    Liked = "#LIKED"
    Dislike = "#DISLIKE"
    Disliked = "#DISLIKED"
    Bookmark = "#BOOKMARK"
    Bookmarked = "#BOOKMARKED"
    Delete = "#DELETE"
    Restore = "#RESTORE"
    Confirm = "#CONFIRM"
    Cancel = "#CANCEL"
    Increase = "#INCREASE"
    Decrease = "#DECREASE"
    Upvote = "#UPVOTE"
    Upvoted = "#UPVOTED"
    Downvote = "#DOWNVOTE"
    Downvoted = "#DOWNVOTED"
    Agree = "#AGREE"
    Agreed = "#AGREED"
    Disagree = "#DISAGREE"
    Disagreed = "#DISAGREED"
    Lock = "#LOCK"
    Unlock = "#UNLOCK"
    Pin = "#PIN"
    Unpin = "#UNPIN"
    Follow = "#FOLLOW"
    Unfollow = "#UNFOLLOW"
    Subscribe = "#SUBSCRIBE"
    Unsubscribe = "#UNSUBSCRIBE"
    Publish = "#PUBLISH"
    Unpublish = "#UNPUBLISH"
    Show = "#SHOW"
    Hide = "#HIDE"
    Allow = "#ALLOW"
    Disallow = "#DISALLOW"
    Approve = "#APPROVE"
    Deny = "#DENY"
    Accept = "#ACCEPT"
    Reject = "#REJECT"
    Star = "#STAR"
    Starred = "#STARRED"
